#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "bmfonts.h"

static const char *item_names[] = {
	"wooden-chest",
	"iron-chest",
	"steel-chest",
	"smart-chest",
	"storage-tank",
	"basic-transport-belt",
	"fast-transport-belt",
	"express-transport-belt",
	"basic-transport-belt-to-ground",
	"fast-transport-belt-to-ground",
	"express-transport-belt-to-ground",
	"basic-splitter",
	"fast-splitter",
	"express-splitter",
	"burner-inserter",

	"basic-inserter",
	"long-handed-inserter",
	"fast-inserter",
	"smart-inserter",
	"small-electric-pole",
	"medium-electric-pole",
	"big-electric-pole",
	"substation",
	"pipe",
	"pipe-to-ground",
	"small-pump",
	"straight-rail",
	"curved-rail",
	"train-stop",
	"rail-signal",

	"rail-chain-signal",
	"diesel-locomotive",
	"cargo-wagon",
	"car",
	"tank",
	"logistic-robot",
	"construction-robot",
	"logistic-chest-active-provider",
	"logistic-chest-passive-provider",
	"logistic-chest-requester",
	"logistic-chest-storage",
	"roboport",
	"copper-cable",
	"red-wire",
	"green-wire",

	"arithmetic-combinator",
	"decider-combinator",
	"constant-combinator",
	"stone-brick",
	"concrete",
	"iron-axe",
	"steel-axe",
	"repair-pack",
	"blueprint",
	"deconstruction-planner",
	"boiler",
	"steam-engine",
	"small-lamp",
	"solar-panel",
	"basic-accumulator",

	"burner-mining-drill",
	"basic-mining-drill",
	"offshore-pump",
	"pumpjack"
};

#define NUM_ITEMS (sizeof item_names / sizeof *item_names)

#define CONSTANT_COMBINATOR_W 1.0
#define DECIDER_COMBINATOR_W 1.5

struct pos2d {
	double x, y;
};


struct glyph_filter_item {
	const char *item_name;
	int value;
};


void create_cell(int glyph_idx, int *entity_id, struct pos2d *pos);

void constantcombinator_create(int filter_id, const struct glyph_filter_item *filter_items, struct pos2d *pos, int prev, int next);
void constantcombinator_createfilter(int filter_id, const struct glyph_filter_item *filter_items);

void decidercombinator_create(int glyph_idx, struct pos2d *pos, int prev, int next);

void small_lamp_create(int item_idx, struct pos2d *pos, int prev, int next);

struct glyph_filter_item *glyph_makefilter(int glyph_idx);


#define GENERATE_ROM
//#define GENERATE_LIGHTS
//#define TEST_GLYPH_FILTER

int main (void)
{
#ifdef GENERATE_ROM
	struct pos2d pos = {};

	int i;
	int entity_id;

	int cell_idx, cell_row;

	printf("do local _ = {\nentities = {\n\n");

	entity_id = 0;
	cell_idx = 0;
	cell_row = 1;

	for (i = 0; i < 256; i++) {
		if (1) { //(isprint(i) && !isspace(i)) || i == 1 || i == 254) {
			create_cell(i, &entity_id, &pos);
			cell_idx++;
			if (cell_idx % 8 == 0) {
				pos.x = 0;
				pos.y += 1;
				if (cell_row % 6 == 0)
					pos.y += 1;
				cell_row++;
			}
		}
	}

	printf("\n},icons={ {index=1,name=\"constant-combinator\"},{index=2,name=\"decider-combinator\"} }\n\n");
	printf("};return _;end\n");
#endif // GENERATE_ROM


#ifdef GENERATE_LIGHTS

	struct pos2d pos = {};

	int i, j;

	printf("do local _ = {\nentities = {\n\n");

	for (i = 0; i < 8; i++) {
		for (j = 0; j < 8; j++) {
			small_lamp_create(i * 8 + j, &pos, j == 0 ? 0 :  i * 8 + j , i * 8 + j + 2);
		}
		pos.x = 0.0;
		pos.y += 1.0;
	}

	printf("\n},icons={ {index=1,name=\"small-lamp\"} }\n\n");
	printf("};return _;end\n");
#endif


#ifdef TEST_GLYPH_FILTER

	// Note: Doesn't free memory :)

	int i;
	for (i = 0; i < 256; i++) {
		if (1) {//isprint(i) && !isspace(i)) {
			glyph_makefilter(i);
		}
	}
	//glyph_makefilter(1);
	//glyph_makefilter(254);
#endif

	return 0;
}

void create_cell(int glyph_idx, int *entity_id, struct pos2d *pos)
{
	int filter_id;
	struct glyph_filter_item *filter_items;

	filter_items = glyph_makefilter(glyph_idx);
	if (!filter_items) {
		perror("Could not allocate filter");
		exit(1);
	}

	/*-----------------------------------------------*/
	/* Create combinators */

	for (filter_id = 0; filter_id < 5; filter_id++) {
		constantcombinator_create(filter_id, filter_items, pos, (filter_id == 0 ? 0 : *entity_id), *entity_id + 2);
		printf(",");
		printf("\n");
		(*entity_id)++;
	}

	// Hack: a) Next item is a decider and the width/pos seems relative to the right-hand side of the entity.
	//       b) Adding the constant combinator has already increased the x position by 1, but to account for
	//          observation (a) another 0.5 has to be added before a decider can be added (deciders are 1.5 wide)
	//
	pos->x += 0.5;

	/*-----------------------------------------------*/
	/* Create decider */

	decidercombinator_create(glyph_idx, pos, *entity_id, 0);
	(*entity_id)++;
	/*-----------------------------------------------*/

	free(filter_items);
}

/*=====================================================================
 * Constant combinators
 *=====================================================================*/

/*
 * prev: prev combinator to connect to
 * next: next combinator to connect to
 */
void constantcombinator_create(int filter_id, const struct glyph_filter_item *filter_items, struct pos2d *pos, int prev, int next)
{
	printf("    {\n");
	printf("        name=\"constant-combinator\",\n");
	printf("        position={x=%.1f,y=%.1f},\n", pos->x, pos->y);
	printf("        connections= { [\"1\"]={");
	if (prev > 0 || next > 0) {
		printf("red={");
		if (prev > 0) {
			printf(" {entity_id=%d}", prev);
			if (next > 0) {
				printf(",");
			}
		}
		if (next > 0) {
			printf(" {entity_id=%d,circuit_id=1}}", next);
		} else {
			printf(" }");
		}
		printf(" }},\n");
	}
	printf("        filters = {\n");
	constantcombinator_createfilter(filter_id, filter_items);
	printf("        }\n");
	printf("    }");

	pos->x += CONSTANT_COMBINATOR_W;
}


void constantcombinator_createfilter(int filter_id, const struct glyph_filter_item *filter_items)
{
	int base, i, item_idx;

	base = filter_id * 15;	/* base index for item names/strings */

	item_idx = 1;

	for (i = 0; i < 15 && base + i < NUM_ITEMS; i++) {
		printf("            {signal={type=\"item\",name=\"%s\"},count=%d,index=%d}",
					filter_items[base + i].item_name,
					filter_items[base + i].value,
					item_idx);
		if (++item_idx % 16 == 0 || base + i == NUM_ITEMS - 1) {
			item_idx = 1;
		} else {
			printf(",");
		}
		printf("\n");
	}
}

/*=====================================================================
 * Decider combinators
 *=====================================================================*/

void decidercombinator_create(int glyph_idx, struct pos2d *pos, int prev, int next)
{
	printf("    {\n");
	printf("        name=\"decider-combinator\",\n");
	printf("        position={x=%.1f,y=%.1f},\n", pos->x, pos->y);
	printf("        direction=2,\n");
	printf("        connections={[\"1\"]={red={{entity_id=%d}}}},\n", prev);
	printf("        conditions={\n");
	printf("            decider={\n");
	printf("                first_signal={type=\"virtual\",name=\"signal-green\"},\n");
	printf("                constant=%d,\n", glyph_idx);
	printf("                comparator=\"=\",\n");
	printf("                output_signal={type=\"virtual\",name=\"signal-everything\"},\n");
	printf("                copy_count_from_input=\"false\"\n");
	printf("            }\n");
	printf("        }\n");
	printf("    },\n");

	pos->x += DECIDER_COMBINATOR_W;
}


/*=====================================================================
 * Small lamp
 *=====================================================================*/

void small_lamp_create(int item_idx, struct pos2d *pos, int prev, int next)
{
	printf("    {\n");
	printf("        name=\"small-lamp\",\n");
	printf("        position={x=%.1f,y=%.1f},\n", pos->x, pos->y);
	printf("        connections= { [\"1\"]={");
	if (prev > 0 || next > 0) {
		printf("red={");
		if (prev > 0) {
			printf(" {entity_id=%d}", prev);
			if (next > 0) {
				printf(",");
			}
		}
		if (next > 0) {
			printf(" {entity_id=%d}}", next);
		} else {
			printf(" }");
		}
		printf(" }},\n");
	}
	printf("        conditions={circuit={first_signal={type=\"item\",name=\"%s\"},constant=1,comparator=\"=\"}}\n", item_names[item_idx]);
	printf("    },\n");

	pos->x += 1.0;
}

/*=====================================================================
 *
 *=====================================================================*/

struct glyph_filter_item *glyph_makefilter(int glyph_idx)
{
	const struct vidfont8 *font = vidfont_getnamedfont(VPU_FONT_8x8_NAME);

	struct glyph_filter_item *filter_items;
	int i;
	int src_pos;
	int bitmask_shift = 7;

	filter_items = malloc(64 * sizeof *filter_items);
	if (!filter_items)
		return NULL;

	src_pos = glyph_idx * 8;

	for (i = 0; i < 64; i++) {
		filter_items[i].item_name = item_names[i];
		filter_items[i].value = (font->pixeldata[src_pos] & (1U << bitmask_shift)) != 0;

#ifdef TEST_GLYPH_FILTER
		//printf("%30s %d\n", filter_items[i].item_name,  filter_items[i].value);
		printf("%c", filter_items[i].value ? '#' : '-');
#endif
		bitmask_shift--;
		if (bitmask_shift < 0) {
			bitmask_shift = 7;
			src_pos++;
#ifdef TEST_GLYPH_FILTER
			printf("\n");
#endif
		}
	}

#ifdef TEST_GLYPH_FILTER
	printf("\n");
#endif

	return filter_items;
}
